#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <libgen.h>

#include <sys/utsname.h>

/*
 * execve wrapping is taken directly from fakeroot
 * please look at fakeroot source code for examples, documentation
 * and overall understanding how this works :-)
 * */


#define LIBCPATH "/scratchbox/host_shared/lib/libc.so.6"

struct next_wrap_st {
	void **doit;
	char *name;
};

void *get_libc()
{
#ifndef RTLD_NEXT
	void *lib=0;
	if (!lib) {
		lib = dlopen(LIBCPATH, RTLD_LAZY);
	}
	if (NULL == lib) {
		fprintf(stderr, "Couldn't find libc at: %s\n", LIBCPATH);
		abort();
	}
	return lib;
#else
	return RTLD_NEXT;
#endif
}


void load_library_symbols(void);

extern int (*next_execve)(const char *file, char *const argv[], char *const envp[]);
int tmp_execve(const char *file, char *const argv[], char *const envp[]) {
	load_library_symbols();
	return next_execve(file,argv,envp);
}

int (*next_execve)(const char *file, char *const argv[], char *const envp[])=tmp_execve;

struct next_wrap_st next_wrap[]= {
	{(void(*))&next_execve, "execve"},
	{NULL, NULL}
};

void load_library_symbols(void)
{
	static int done=0;
	int i;
	char *msg;

	if (!done) {
		for (i=0; next_wrap[i].doit; i++) {
			*(next_wrap[i].doit)=dlsym(get_libc(), next_wrap[i].name);
			if ( (msg = dlerror()) != NULL) {
				fprintf (stderr, "dlsym(%s): %s\n", next_wrap[i].name, msg);
			}
		}
	}
}


int execve (file, argv, envp)
	const char *file;
	char *const argv[];
	char *const envp[];
{
	char *dir, *bin, *new_file, *redir_list, *li, **my_argv, **tmp, **tmp2;
	char *tmp_file;
	char buf[1024]; // hash bang detection buffer
	char interpreter[1024];
	char interpreter_arg[1024];

	int c=0,i=0,j=0,hash_bang=0;
	FILE *fp;
	
	//fprintf(stderr, "fake execve!\n");

	if (!(redir_list = getenv("SBOX_REDIRECT_BINARIES"))) {
		/* default as there is nothing to redirect */
		return next_execve(file, argv, envp);
	}

	// true magic custom hash-bang handler!
	
	memset(buf, 0, 1024);
	memset(interpreter, 0, 1024);
	
	if ((fp = fopen(file, "r"))) {
		c = fread(buf, 1023, 1, fp);
		if (buf[0] == '#' && buf[1] == '!') {
			// we have hash bang
			//fprintf(stderr, "we have a hash bang!\n");
			i = 2;
			while ((buf[i] == ' ' || buf[i] == '\t') && buf[i] != '\0' && buf[i] != '\n') i++;
			j = i;
			while (buf[j] != ' ' && buf[j] != '\t' && buf[j] != '\0' && buf[j] != '\n') j++;
			strncpy(interpreter, &buf[i], j - i);
			while ((buf[j] == ' ' || buf[j] == '\t') && buf[j] != '\0' && buf[j] != '\n') j++;
			i = j;
			while (buf[j] != '\n' && buf[j] != '\0') j++;
			strncpy(interpreter_arg, &buf[i], j - i);
			//fprintf(stderr, "interpreter: [%s]\n", interpreter);
			//fprintf(stderr, "arg: [%s]\n", interpreter_arg);
			hash_bang = 1;
		}
		fclose(fp);
	}

	

	if (hash_bang) {
		tmp_file = interpreter;
		tmp = (char **)argv;
		c = 0;
		while (*tmp) {
			c++;
			tmp++;
		}
		my_argv = malloc(sizeof(char **)* (c + 3));
		tmp = (char **)argv;
		tmp2 = my_argv;
		*tmp2++ = (char *)file;
		if (interpreter_arg[0] != '\0') {
			*tmp2++ = interpreter_arg;
		}

		*tmp2++ = (char *)file;
		
		tmp++; // skip argv[0]
		while (*tmp) {
			*tmp2++ = *tmp++;
		}
		*tmp2 = NULL;
	} else {
		my_argv = (char **)argv;
		tmp_file = (char *)file;
	}

	dir = dirname(strdup(tmp_file));
	bin = basename(strdup(tmp_file));

	while ((li = strsep(&redir_list, ":"))) {
		if ( (strcmp(tmp_file, li) == 0) ) {
			new_file = malloc((strlen("/bin") + strlen(bin) + 2)*sizeof(char));
			new_file[0] = '\0';
			strcat(new_file, "/bin");
			strcat(new_file, "/");
			strcat(new_file, bin);
			return next_execve(new_file, my_argv, envp);
		}
	}

	/* fall back to default behaviour */
	
	//fprintf(stderr, "default handler\n");
	return next_execve(file, argv, envp);
}

int uname(struct utsname *buf) 
{
	char *my_sysname, *my_nodename, *my_release, *my_version, *my_machine;
	if ((my_sysname = getenv("SBOX_UNAME_SYSNAME"))) {
		strncpy(buf->sysname, my_sysname, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->sysname,"Linux", _UTSNAME_LENGTH);
	}

	if ((my_nodename = getenv("SBOX_UNAME_NODENAME"))) {
		strncpy(buf->nodename, my_nodename, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->nodename, "fake-node", _UTSNAME_LENGTH);
	}

	if ((my_release = getenv("SBOX_UNAME_RELEASE"))) {
		strncpy(buf->release, my_release, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->release, "2.4.21", _UTSNAME_LENGTH);
	}

	if ((my_version = getenv("SBOX_UNAME_VERSION"))) {
		strncpy(buf->version, my_version, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->version, "#1 Sun Jul 20 19:23:36 EST 2003", _UTSNAME_LENGTH);
	}

	if ((my_machine = getenv("SBOX_UNAME_MACHINE"))) {
		strncpy(buf->machine, my_machine, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->machine, "i386", _UTSNAME_LENGTH);
	}
	return 0;
}
